from datetime import datetime, timedelta
import pandas as pd
import numpy as np
from hbshare.quant.CChen.cta_factor.factor_func import factor_compute, get_stock_data, get_mr_data
from hbshare.quant.CChen.stk import trade_calendar
from hbshare.quant.CChen.cta_factor.const import stock_data_indicator
from hbshare.quant.CChen.cta_factor.factor_algo import (
    tscarry, carry, tsmom, tsrev, xsmom, xsrev, tswr, xswr, mr, mrchg, tsmr,
    tsbasismom, xsbasismom, tspoichg, xspoichg, xssigma,
    xsmr, xstsmr, xsmrchg, tspvolchg, xspvolchg, tsskew, xsskew, xsmr2, xsmrchg2
    # tsstock, xsstock
)


def run(sql_path, sql_info):
    start_date = datetime(2010, 1, 1).date()
    # end_date = datetime(2021, 4, 9).date()
    end_date = datetime.now().date()

    window_days_list = [1, 2, 3, 5, 10, 20, 30, 40, 50, 60, 90, 100, 150, 200, 250]
    liq_d = 20

    table = 'hsjy_fut_com_index'
    table_wr = 'hsjy_fut_wr'
    table_mr = 'hsjy_fut_memberrank'
    table_contracts_info = 'hsjy_fut_info_c'
    table_p_info = 'hsjy_fut_info_p'
    table_raw = 'hsjy_fut_com'

    data = pd.read_sql_query(
        'select * from ' + table + ' where TDATE<='
        + end_date.strftime('%Y%m%d')
        + ' and TDATE>=' + (start_date - timedelta(days=500)).strftime('%Y%m%d')
        + ' order by TDATE',
        sql_path
    )
    col_adj = ['VOL', 'AMT', 'OI', 'PVOL', 'POI', 'PAMT']
    date_2020 = data.index[data['TDATE'] < datetime(2020, 1, 1).date()]
    data.loc[date_2020, col_adj] = data.loc[date_2020, col_adj] / 2
    data_wr0 = pd.read_sql_query(
        'select TDATE, EXCHANGE, PCODE, PNAME, WRQCURRENT from ' + table_wr + ' where TDATE<='
        + end_date.strftime('%Y%m%d')
        + ' and TDATE>=' + (start_date - timedelta(days=500)).strftime('%Y%m%d')
        + ' and FREQ=5 order by TDATE, EXCHANGE, PCODE',
        sql_path
    ).drop_duplicates(subset=['TDATE', 'EXCHANGE', 'PCODE']).reset_index(drop=True)
    calendar = data[['TDATE']].sort_values(by='TDATE').drop_duplicates().reset_index(drop=True)
    data_wr = pd.DataFrame()
    for e in data['EXCHANGE'].drop_duplicates().tolist():
        PCODEs = data[data['EXCHANGE'] == e]['PCODE'].drop_duplicates().tolist()
        for p in PCODEs:
            data_wr_p = data_wr0[np.array(data_wr0['EXCHANGE'] == e) & np.array(data_wr0['PCODE'] == p)]
            data_wr_p = pd.merge(calendar, data_wr_p, on='TDATE', how='left').fillna(method='ffill')
            data_wr_p = data_wr_p[data_wr_p['WRQCURRENT'] >= 0].reset_index(drop=True)
            data_wr = pd.concat([data_wr, data_wr_p])
    data_wr = data_wr.reset_index(drop=True).rename(columns={'WRQCURRENT': 'x'})

    data_mr = get_mr_data(start_date=start_date, end_date=end_date, path=sql_path, table=table_mr)
    # data_mr = data_mr[data_mr['SIDE'] != 1]
    data_mr['SIDE'] = data_mr['SIDE'].apply(lambda x: 1 if x == 3 else (-1 if x == 4 else 0))
    data_mr['MR'] = data_mr['MR'] * data_mr['SIDE']
    data_mr['MRCHG'] = data_mr['MRCHG'] * data_mr['SIDE']
    data_info = pd.read_sql_query(
        'select * from ' + table_contracts_info, sql_path
    )
    data_mr = pd.merge(data_mr, data_info[['CCODE', 'EXCHANGE', 'PCODE']], on='CCODE', how='left')
    data_mr = data_mr.groupby(by=['TDATE', 'SIDE', 'EXCHANGE', 'PCODE']).sum().reset_index()

    data_contracts = pd.read_sql_query(
        'select * from ' + table_raw + ' where CCODE in ' + str(tuple(np.unique(data[['CCODE', 'CCODE2']].dropna())))
        + ' order by CCODE, TDATE',
        sql_path
    ).drop_duplicates(subset=['TDATE', 'CCODE'])

    # 库存数据
    data_stock = get_stock_data(
        data_indicator=stock_data_indicator, path=sql_path, table_p=table_p_info
    ).rename(columns={'WRQCURRENT': 'x'})

    factor_compute(
        window_days_list=[10, 20, 30, 50, 90, 150, 250],
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xssigma,
        liq_days=liq_d,
        data_raw=data_contracts,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list[:10],
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xsbasismom,
        liq_days=liq_d,
        data_raw=data_contracts,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list[:10],
        data=data,
        factor_func=tsbasismom,
        liq_days=liq_d,
        data_raw=data_contracts,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=[20, 50],
        hedge_ratio_list=[50, 60, 70, 80, 90],
        data=data,
        factor_func=carry,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        factor_func=tsmom,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        factor_func=tscarry,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list[2:],
        data=data, factor_func=tsskew,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        factor_func=tsrev,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xsmom,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list[2:],
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xsskew,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xsrev,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        factor_func=tspoichg,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        factor_func=tspvolchg,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xspoichg,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        factor_func=xspvolchg,
        liq_days=liq_d,
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_wr=data_wr,
        factor_func=xswr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        data_wr=data_wr,
        factor_func=tswr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_wr=data_stock,
        factor_func=xswr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path,
        name='stock',
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        data_wr=data_stock,
        factor_func=tswr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path,
        name='stock',
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        data_mr=data_mr,
        factor_func=mr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info, to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_mr=data_mr,
        factor_func=xsmr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info, to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_mr=data_mr,
        factor_func=xsmr2,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info, to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        data_mr=data_mr,
        factor_func=mrchg,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_mr=data_mr,
        factor_func=xsmrchg,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_mr=data_mr,
        factor_func=xsmrchg2,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        data=data,
        data_mr=data_mr,
        factor_func=tsmr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info,
        to_sql_path=sql_path
    )

    factor_compute(
        window_days_list=window_days_list,
        hedge_ratio_list=[50, 75],
        data=data,
        data_mr=data_mr,
        factor_func=xstsmr,
        liq_days=liq_d,
        price='OPEN',
        sql_info=sql_info, to_sql_path=sql_path
    )


if __name__ == '__main__':
    from hbshare.quant.CChen.db_const import sql_write_path_work, sql_user_work
    run(
        sql_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )




"""
alpha标的sourcing
"""
import hbshare as hbs
import pandas as pd
import numpy as np
from tqdm import tqdm
import Levenshtein
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql, get_trading_day_list
from Arbitrage_backtest import cal_annual_return, cal_annual_volatility, cal_sharpe_ratio, cal_max_drawdown
import jieba
import jieba.posseg as pseg
import numpy as np
from sklearn.cluster import KMeans
from sklearn import feature_extraction
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.cluster import Birch
import matplotlib.pyplot as plt

plt.style.use('seaborn')


## 统计数量
# sql_script = "SELECT count(*) FROM st_hedge.t_st_jjxx"
# data = pd.DataFrame(hbs.db_data_query('highuser', sql_script)['data'])
# print(data)


class AlphaSourcing:
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date

    def _get_fund_info(self):
        sql_script = \
            "SELECT clrq, jjdm, jjmc, jjjc FROM st_hedge.t_st_jjxx where cpfl = '4' and ejfl = '1002' and jjzt = '0'"
        data = pd.DataFrame(hbs.db_data_query('highuser', sql_script, page_size=50000)['data'])
        # 剔除300
        data = data[~data['jjmc'].str.contains('300')]
        # 剔除1000
        data = data[~data['jjmc'].str.contains('1000')]
        # 剔除中性
        data = data[~data['jjmc'].str.contains('对冲')]
        data = data[~data['jjmc'].str.contains('中性')]
        # 成立满一年
        data = data[data['clrq'] < self.start_date]

        fund_dict = data.set_index('jjmc')['jjdm'].to_dict()

        return fund_dict

    @staticmethod
    def _get_fund_nav_from_sql(start_date, end_date, fund_dict):
        """
        获取db的私募基金净值数据
        """
        nav_series_list = []
        for name, fund_id in tqdm(fund_dict.items()):
            sql_script = "SELECT a.jjdm fund_id, b.jzrq TRADEDATE, b.fqdwjz as ADJ_NAV from " \
                         "st_hedge.t_st_jjxx a, st_hedge.t_st_rhb b where a.cpfl = '4' and a.jjdm = b.jjdm " \
                         "and a.jjzt not in ('3') " \
                         "and a.jjdm = '{}' and b.jzrq >= {} and b.jzrq <= {} " \
                         "order by b.jzrq".format(fund_id, start_date, end_date)
            res = hbs.db_data_query("highuser", sql_script, page_size=5000)
            if len(res['data']) != 0:
                data = pd.DataFrame(res['data']).set_index('TRADEDATE')['ADJ_NAV']
                data.name = name
                nav_series_list.append(data)
            else:
                pass
        df = pd.concat(nav_series_list, axis=1).sort_index()

        return df

    @staticmethod
    def lev_metric(x, y):
        distance = Levenshtein.distance(x, y)
        return 1 - distance / (len(x) + len(y))

    def run(self):
        fund_dict = self._get_fund_info()
        nav_df = self._get_fund_nav_from_sql(self.start_date, self.end_date, fund_dict)
        trading_day_list = get_trading_day_list(self.start_date, self.end_date, frequency="week")
        trading_day_list = [x for x in trading_day_list if x != "20210930"]
        nav_df = nav_df.reindex(trading_day_list)
        # 数据缺失剔除
        count_df = nav_df.isnull().sum()
        selected_list = count_df[count_df <= 3].index.tolist()
        nav_df = nav_df[selected_list]
        # 均值填充
        return_df = nav_df.pct_change(limit=1)
        return_df.iloc[0] = 0.
        return_df = return_df.T
        return_df = return_df.fillna(return_df.mean()).T
        nav_df = (1 + return_df).cumprod()
        # 跟踪误差校验：10%为界
        sql_script = "SELECT JYRQ as TRADEDATE, ZQDM, SPJG as TCLOSE from funddb.ZSJY WHERE" \
                     " ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format("000905", self.start_date, self.end_date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        data = pd.DataFrame(res['data'])
        benchmark_series = data.set_index('TRADEDATE')['TCLOSE'].reindex(trading_day_list)
        return_df = nav_df.pct_change().dropna().sub(benchmark_series.pct_change().dropna(), axis=0).sort_index()
        annual_vol = return_df.std() * np.sqrt(52)
        vol_list = annual_vol[annual_vol < 0.15].index.tolist()  # 以幻方为界
        return_df = return_df[vol_list]
        excess_df = (1 + return_df).cumprod()
        # 剔除区间内超额的标的
        period_excess = excess_df.iloc[-1] / excess_df.iloc[0] - 1
        selected_list = period_excess[period_excess >= 0.05].index.tolist()
        excess_df = excess_df[selected_list]
        "========================================================================================================"
        # 文本聚类
        stop_words = ['私募', '基金', '证券', '投资', '集合', '资金', '计划', '号', '期', '\n']+[str(i) for i in range(1, 500)]
        jieba.add_word("中证500")
        jieba.add_word("500指数")
        title_list = excess_df.columns.tolist()
        corpus = [" ".join([w for w in jieba.lcut(str(x)) if w not in stop_words]) for x in title_list]
        # title_dict = {idx: title for idx, title in enumerate(title_list)}
        #将文本中的词语转换为词频矩阵 矩阵元素a[i][j] 表示j词在i类文本下的词频
        vectorizer = CountVectorizer()
        #该类会统计每个词语的tf-idf权值
        transformer = TfidfTransformer()
        #第一个fit_transform是计算tf-idf 第二个fit_transform是将文本转为词频矩阵
        tfidf = transformer.fit_transform(vectorizer.fit_transform(corpus))
        #获取词袋模型中的所有词语
        # word = vectorizer.get_feature_names()
        #将tf-idf矩阵抽取出来，元素w[i][j]表示j词在i类文本中的tf-idf权重
        weight = tfidf.toarray()

        print('start cluster Birch -------------------')
        cluster = Birch(threshold=0.3, n_clusters=None)
        cluster.fit_predict(weight)

        # cluster_dict = {}
        # # cluster_dict key为Birch聚类后的每个类，value为 title对应的index
        # for index, value in enumerate(cluster.labels_):
        #     if value not in cluster_dict:
        #         cluster_dict[value] = [index]
        #     else:
        #         cluster_dict[value].append(index)
        # print(cluster_dict)

        label_df = pd.DataFrame({"fund_name": title_list, "label": cluster.labels_}).sort_values(by='label')
        distinct_list = label_df.drop_duplicates(subset='label', keep='first')['fund_name'].tolist()
        distinct_return_df = return_df[distinct_list]

        "==================================================================================================="
        dis_nav_df = (1 + distinct_return_df).cumprod()

        # 净值指标
        performance_df = pd.DataFrame(
            index=distinct_return_df.columns, columns=['超额年化收益', '超额年化波动', '最大回撤', 'Sharpe', '区间收益9-2', '今年以来超额'])
        performance_df.loc[:, '超额年化收益'] = distinct_return_df.apply(cal_annual_return, axis=0)
        performance_df.loc[:, '超额年化波动'] = distinct_return_df.apply(cal_annual_volatility, axis=0)
        performance_df.loc[:, '最大回撤'] = dis_nav_df.apply(cal_max_drawdown, axis=0)
        performance_df.loc[:, 'Sharpe'] = distinct_return_df.apply(lambda x: cal_sharpe_ratio(x, 0.015), axis=0)
        performance_df.loc[:, '区间收益9-2'] = dis_nav_df.loc["20220211"] / dis_nav_df.loc["20210903"] - 1
        performance_df.loc[:, '今年以来超额'] = dis_nav_df.iloc[-1] / dis_nav_df.loc["20211231"] - 1

        # 超额稳定性聚类
        kmeans = KMeans(n_clusters=2)
        df = performance_df[['最大回撤', 'Sharpe', '区间收益9-2']]
        result = kmeans.fit(df)
        df['label'] = result.labels_

        # cluster_center = pd.DataFrame(columns=['最大回撤', 'Sharpe', '区间收益9-2'], data=result.cluster_centers_)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(df[df['label'] == 1]['最大回撤'],
                   df[df['label'] == 1]['区间收益9-2'],
                   df[df['label'] == 1]['Sharpe'], c='r')
        ax.scatter(df[df['label'] == 0]['最大回撤'],
                   df[df['label'] == 0]['区间收益9-2'],
                   df[df['label'] == 0]['Sharpe'], c='g')
        ax.set_xlabel('max_drawdown')
        ax.set_ylabel('9-2_period_ret')
        ax.set_zlabel('Sharpe')

        stable_list = df[df['label'] == 1].index.tolist()
        unstable_list = df[df['label'] == 0].index.tolist()
        (1 + distinct_return_df[stable_list].mean(axis=1)).cumprod().to_clipboard()
        (1 + distinct_return_df[unstable_list].mean(axis=1)).cumprod().to_clipboard()


if __name__ == '__main__':
    s_date = '20210827'
    e_date = '20220916'

    AlphaSourcing(s_date, e_date).run()


# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
from hbshare.fe.xwq.analysis.orm.hbdb import HBDB
import numpy as np
import pandas as pd
import os

from WindPy import w
w.start()  # 默认命令超时时间为120秒，如需设置超时时间可以加入waitTime参数，例如waitTime=60,即设置命令超时时间为60秒
w.isconnected()  # 判断WindPy是否已经登录成功

from matplotlib.ticker import FuncFormatter
import seaborn as sns
import matplotlib.pyplot as plt
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False
line_color_list = ['#F04950', '#6268A2', '#959595', '#333335', '#EE703F', '#7E4A9B', '#8A662C', '#44488E', '#BA67E9', '#3FAEEE',
                   '#DC494F', '#9EA2C9', '#CBCBCB', '#5A5A63', '#F8B399', '#B074D1', '#C8A060', '#1C26CE', '#CB96E9', '#AFDEFA']


def to_100percent(temp, position):
    return '%1.0f'%(temp * 100) + '%'


def get_date(start_date, end_date):
    calendar_df = HBDB().read_cal(start_date, end_date)
    calendar_df = calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
    calendar_df['CALENDAR_DATE'] = calendar_df['CALENDAR_DATE'].astype(str)
    calendar_df = calendar_df.sort_values('CALENDAR_DATE')
    calendar_df['IS_OPEN'] = calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
    calendar_df['YEAR_MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
    calendar_df['MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:6])
    calendar_df['MONTH_DAY'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
    calendar_df = calendar_df[(calendar_df['CALENDAR_DATE'] >= start_date) & (calendar_df['CALENDAR_DATE'] <= end_date)]
    trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    trade_df = trade_df[(trade_df['TRADE_DATE'] >= start_date) & (trade_df['TRADE_DATE'] <= end_date)]
    report_df = calendar_df.drop_duplicates('YEAR_MONTH', keep='last').rename(columns={'CALENDAR_DATE': 'REPORT_DATE'})
    report_df = report_df[report_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231'])]
    report_df = report_df[(report_df['REPORT_DATE'] >= start_date) & (report_df['REPORT_DATE'] <= end_date)]
    report_trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    report_trade_df = report_trade_df.sort_values('TRADE_DATE').drop_duplicates('YEAR_MONTH', keep='last')
    report_trade_df = report_trade_df[report_trade_df['MONTH'].isin(['03', '06', '09', '12'])]
    report_trade_df = report_trade_df[(report_trade_df['TRADE_DATE'] >= start_date) & (report_trade_df['TRADE_DATE'] <= end_date)]
    calendar_trade_df = calendar_df[['CALENDAR_DATE']].merge(trade_df[['TRADE_DATE']], left_on=['CALENDAR_DATE'], right_on=['TRADE_DATE'], how='left')
    calendar_trade_df['TRADE_DATE'] = calendar_trade_df['TRADE_DATE'].fillna(method='ffill')
    calendar_trade_df = calendar_trade_df[(calendar_trade_df['TRADE_DATE'] >= start_date) & (calendar_trade_df['TRADE_DATE'] <= end_date)]
    return calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df

class Crowding:
    def __init__(self, index, index_name_dic, start_date, end_date, data_path):
        self.index = index
        self.start_date = start_date
        self.end_date = end_date
        self.start_date_hyphen = datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
        self.end_date_hyphen = datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
        self.data_path = data_path
        self.index_name_dic = index_name_dic
        self.load()

    def load(self):
        self.calendar_df, self.report_df, self.trade_df, self.report_trade_df, self.calendar_trade_df = get_date((datetime.strptime(self.start_date, '%Y%m%d') - timedelta(20)).strftime('%Y%m%d'), self.end_date)

        # 指数收盘点位、成交额、换手率、主动资金净流入
        self.index_data = w.wsd(self.index, "close,amt,turn,mfd_inflow_m", self.start_date_hyphen, self.end_date_hyphen, "unit=1", usedf=True)[1].reset_index()
        self.index_data.to_hdf('{0}{1}.hdf'.format(self.data_path, self.index), key='table', mode='w')
        self.index_data = pd.read_hdf('{0}{1}.hdf'.format(self.data_path, self.index), key='table')
        self.index_data.columns = ['TRADE_DATE', 'CLOSE_INDEX', 'TURNOVER_VALUE', 'TURNOVER_RATE', 'MF_NET_INFLOW']
        self.index_data['INDEX_SYMBOL'] = self.index
        self.index_data['TRADE_DATE'] = self.index_data['TRADE_DATE'].apply(lambda x: str(x).replace('-', ''))
        self.index_data = self.index_data[['INDEX_SYMBOL', 'TRADE_DATE', 'CLOSE_INDEX', 'TURNOVER_VALUE', 'TURNOVER_RATE', 'MF_NET_INFLOW']]
        self.index_data = self.index_data.sort_values('TRADE_DATE')

        # 市场收盘点位、成交额、换手率、主动资金净流入
        self.index_market_data = w.wsd("881001.WI", "close,amt,turn,mfd_inflow_m", self.start_date_hyphen, self.end_date_hyphen, "unit=1", usedf=True)[1].reset_index()
        self.index_market_data.to_hdf('{0}{1}.hdf'.format(self.data_path, '881001.WI'), key='table', mode='w')
        self.index_market_data = pd.read_hdf('{0}{1}.hdf'.format(self.data_path, '881001.WI'), key='table')
        self.index_market_data.columns = ['TRADE_DATE', 'CLOSE_INDEX', 'TURNOVER_VALUE', 'TURNOVER_RATE', 'MF_NET_INFLOW']
        self.index_market_data['INDEX_SYMBOL'] = '881001.WI'
        self.index_market_data['TRADE_DATE'] = self.index_market_data['TRADE_DATE'].apply(lambda x: str(x).replace('-', ''))
        self.index_market_data = self.index_market_data[['INDEX_SYMBOL', 'TRADE_DATE', 'CLOSE_INDEX', 'TURNOVER_VALUE', 'TURNOVER_RATE', 'MF_NET_INFLOW']]
        self.index_market_data = self.index_market_data.sort_values('TRADE_DATE')

        # 指数成分股（每季度更新）
        index_cons_list = []
        for date in self.report_trade_df['TRADE_DATE'].unique().tolist():
            date_hyphen = datetime.strptime(date, '%Y%m%d').strftime('%Y-%m-%d')
            index_cons_date = w.wset("sectorconstituent", "date={0};windcode={1}".format(date_hyphen, self.index), usedf=True)[1]
            index_cons_list.append(index_cons_date)
        self.index_cons = pd.concat(index_cons_list)
        self.index_cons.to_hdf('{0}{1}_cons.hdf'.format(self.data_path, self.index), key='table', mode='w')
        self.index_cons = pd.read_hdf('{0}{1}_cons.hdf'.format(self.data_path, self.index), key='table')
        self.index_cons.columns = ['TRADE_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME']
        self.index_cons['INDEX_SYMBOL'] = self.index
        self.index_cons['TRADE_DATE'] = self.index_cons['TRADE_DATE'].apply(lambda x: str(x.date()).replace('-', ''))
        self.index_cons['TICKER_SYMBOL'] = self.index_cons['TICKER_SYMBOL'].apply(lambda x: x.split('.')[0])
        self.index_cons = self.index_cons[['INDEX_SYMBOL', 'TRADE_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME']]
        self.index_cons = self.index_cons.sort_values(['TRADE_DATE', 'TICKER_SYMBOL'])
        self.index_cons['IS_CON'] = 1
        self.index_cons = self.index_cons.pivot(index='TRADE_DATE', columns='TICKER_SYMBOL', values='IS_CON').sort_index().fillna(0.0)
        self.index_cons = self.trade_df[['TRADE_DATE']].set_index('TRADE_DATE').merge(self.index_cons, left_index=True, right_index=True, how='left')
        self.index_cons = self.index_cons.fillna(method='ffill').dropna()

        # 股票行情
        stock_daily_k_path = '{0}stock_daily_k.hdf'.format(self.data_path)
        if os.path.isfile(stock_daily_k_path):
            existed_stock_daily_k = pd.read_hdf(stock_daily_k_path, key='table')
            max_date = max(existed_stock_daily_k['TRADE_DATE'])
            start_date = max(str(max_date), '20150101')
        else:
            existed_stock_daily_k = pd.DataFrame()
            start_date = '20150101'
        trade_df = self.trade_df[(self.trade_df['TRADE_DATE'] > start_date) & (self.trade_df['TRADE_DATE'] <= self.end_date)]
        stock_daily_k_list = []
        for date in trade_df['TRADE_DATE'].unique().tolist():
            stock_daily_k_date = HBDB().read_stock_daily_k_given_date(date)
            stock_daily_k_date = stock_daily_k_date[['TRADE_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'CLOSE_PRICE']] if len(stock_daily_k_date) > 0 else pd.DataFrame(columns=['TRADE_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'CLOSE_PRICE'])
            star_stock_daily_k_date = HBDB().read_star_stock_daily_k_given_date(date)
            star_stock_daily_k_date = star_stock_daily_k_date[['TRADE_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'CLOSE_PRICE']] if len(star_stock_daily_k_date) > 0 else pd.DataFrame(columns=['TRADE_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'CLOSE_PRICE'])
            stock_daily_k_date = pd.concat([stock_daily_k_date, star_stock_daily_k_date])
            stock_daily_k_list.append(stock_daily_k_date)
            print(date)
        self.stock_daily_k = pd.concat([existed_stock_daily_k] + stock_daily_k_list, ignore_index=True)
        self.stock_daily_k = self.stock_daily_k.sort_values(['TRADE_DATE', 'TICKER_SYMBOL'])
        self.stock_daily_k = self.stock_daily_k.reset_index().drop('index', axis=1)
        self.stock_daily_k.to_hdf(stock_daily_k_path, key='table', mode='w')
        self.stock_daily_k = pd.read_hdf(stock_daily_k_path, key='table')
        self.stock_daily_k = self.stock_daily_k.sort_values(['TRADE_DATE', 'TICKER_SYMBOL'])
        self.stock_daily_k = self.stock_daily_k.pivot(index='TRADE_DATE', columns='TICKER_SYMBOL', values='CLOSE_PRICE').sort_index()
        self.stock_daily_k_max = self.stock_daily_k.rolling(60).max()
        self.stock_new_high = (self.stock_daily_k >= self.stock_daily_k_max).astype(int)
        self.stock_daily_k_mean = self.stock_daily_k.rolling(20).mean()
        self.stock_mean_above = (self.stock_daily_k > self.stock_daily_k_mean).astype(int)

    def quantile_definition(self, col, idxs, daily_df):
        part_df = daily_df.iloc[list(map(int, idxs))].copy(deep=True)
        q = (1.0 - np.count_nonzero(part_df[col].iloc[-1] <= part_df[col]) / len(part_df))
        return q

    def get_crowding(self):
        index_data = self.index_data.merge(self.index_market_data[['TRADE_DATE', 'TURNOVER_VALUE']].rename(columns={'TURNOVER_VALUE': 'TURNOVER_VALUE_MARKET'}), on=['TRADE_DATE'], how='left')
        index_data['TURNOVER_VALUE_RATE'] = index_data['TURNOVER_VALUE'] / index_data['TURNOVER_VALUE_MARKET']
        index_data['MF_NET_INFLOW_RATE'] = index_data['MF_NET_INFLOW'] / index_data['TURNOVER_VALUE']
        stock_new_high = (self.stock_new_high * self.index_cons).sum(axis=1) / self.index_cons.sum(axis=1)
        stock_new_high = pd.DataFrame(stock_new_high, columns=['NEW_HIGH']).reset_index()
        stock_mean_above = (self.stock_mean_above * self.index_cons).sum(axis=1) / self.index_cons.sum(axis=1)
        stock_mean_above = pd.DataFrame(stock_mean_above, columns=['MEAN_ABOVE']).reset_index()
        index_data = index_data.merge(stock_new_high, on=['TRADE_DATE'], how='left').merge(stock_mean_above, on=['TRADE_DATE'], how='left')
        index_data = index_data[['INDEX_SYMBOL', 'TRADE_DATE', 'CLOSE_INDEX', 'TURNOVER_VALUE_RATE', 'TURNOVER_RATE', 'MF_NET_INFLOW_RATE', 'NEW_HIGH', 'MEAN_ABOVE']]
        index_data = index_data.sort_values('TRADE_DATE')
        index_data['IDX'] = range(len(index_data))
        for col in list(index_data.columns)[3:]:
            index_data[col] = index_data['IDX'].rolling(window=250, min_periods=250, center=False).apply(lambda x: self.quantile_definition(col, x, index_data))
        index_data['CROWDING'] = index_data[list(index_data.columns)[3:]].mean(axis=1)
        index_data = index_data.dropna(subset=['CROWDING'])
        index_data['CROWDING_MEAN'] = index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).mean()
        index_data['CROWDING_UP1'] = index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).mean() + 1.0 * index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).std(ddof=1)
        index_data['CROWDING_DOWN1'] = index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).mean() - 1.0 * index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).std(ddof=1)
        index_data['CROWDING_UP15'] = index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).mean() + 1.5 * index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).std(ddof=1)
        index_data['CROWDING_DOWN15'] = index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).mean() - 1.5 * index_data['CROWDING'].rolling(window=250, min_periods=1, center=False).std(ddof=1)

        index_data_disp = index_data.merge(self.trade_df[['TRADE_DATE', 'IS_WEEK_END']], on=['TRADE_DATE'], how='left')
        index_data_disp = index_data_disp[index_data_disp['IS_WEEK_END'] == '1']
        index_data_disp = index_data_disp[index_data_disp['TRADE_DATE'] >= '20200101']
        index_data_disp['TRADE_DATE'] = index_data_disp['TRADE_DATE'].apply(lambda x: datetime.strptime(x, '%Y%m%d'))
        fig, ax = plt.subplots(figsize=(12, 6))
        ax1 = ax
        ax2 = ax.twinx()
        ax1.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CROWDING'].values, '-', linewidth=3.0, label='拥挤度指标', color=line_color_list[1])
        ax1.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CROWDING_MEAN'].values, '--', label='均值', color=line_color_list[2])
        ax1.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CROWDING_UP1'].values, '--', label='均值±1.0*标准差', color=line_color_list[7])
        ax1.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CROWDING_UP15'].values, '--', label='均值±1.5*标准差', color=line_color_list[3])
        ax1.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CROWDING_DOWN1'].values, '--', color=line_color_list[7])
        ax1.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CROWDING_DOWN15'].values, '--', color=line_color_list[3])
        ax2.plot(index_data_disp['TRADE_DATE'].values, index_data_disp['CLOSE_INDEX'].values, '-', linewidth=3.0, label=self.index_name_dic[self.index] + '（右轴）', color=line_color_list[0])
        h1, l1 = ax1.get_legend_handles_labels()
        h2, l2 = ax2.get_legend_handles_labels()
        plt.legend(handles=h1 + h2, labels=l1 + l2, loc=8, bbox_to_anchor=(0.5, -0.15), ncol=5)
        ax1.set_ylim([0, 1.3])
        ax1.yaxis.set_major_formatter(FuncFormatter(to_100percent))
        ax1.set_xlabel('')
        ax2.set_xlabel('')
        ax1.set_ylabel('')
        ax2.set_ylabel('')
        plt.title('{0}拥挤度'.format(self.index_name_dic[self.index]))
        plt.tight_layout()
        sns.despine(top=True, right=False, left=False, bottom=False)
        plt.savefig('{0}{1}.png'.format(self.data_path, self.index_name_dic[self.index]))
        return index_data


if __name__ == '__main__':
    start_date = '20180101'
    end_date = '20220930'
    data_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/crowding/'
    index_list = ['8841448.WI', '884963.WI', '884045.WI', '8841269.WI', '886032.WI', '884878.WI', '8841278.WI', '801740.SI']
    index_name_dic = {'8841448.WI': '先进制造', '884963.WI': '动力电池', '884045.WI': '光伏', '8841269.WI': '新能源整车',
                      '886032.WI': '汽车零部件', '884878.WI': '半导体', '8841278.WI': '消费电子', '801740.SI': '军工'}
    index_crowding_list = []
    for index in index_list:
        index_data = Crowding(index, index_name_dic, start_date, end_date, data_path).get_crowding()
        index_crowding_list.append(index_data)
    index_crowding = pd.concat(index_crowding_list)
    index_crowding.to_hdf('{0}index_crowding.hdf'.format(data_path), key='table', mode='w')
    index_crowding = pd.read_hdf('{0}index_crowding.hdf'.format(data_path), key='table')
    index_crowding = index_crowding.pivot(index='TRADE_DATE', columns='INDEX_SYMBOL', values='CROWDING')
    index_crowding_up = index_crowding.rolling(window=250, min_periods=1, center=False).mean() + 1.0 * index_crowding.rolling(window=250, min_periods=1, center=False).std(ddof=1)
    index_crowding_down = index_crowding.rolling(window=250, min_periods=1, center=False).mean() - 1.0 * index_crowding.rolling(window=250, min_periods=1, center=False).std(ddof=1)
    index_relative_crowding = (index_crowding - index_crowding_down) / (index_crowding_up - index_crowding_down)
    index_relative_crowding_change = index_relative_crowding.pct_change(20)

    plt.figure(figsize=(6, 6))
    plt.scatter(index_relative_crowding.iloc[-1], index_relative_crowding_change.iloc[-1], color=line_color_list[1])
    for i in range(len(index_relative_crowding.iloc[-1])):
        plt.annotate([index_name_dic[index] for index in list(index_relative_crowding.iloc[-1].index)][i], xy=(index_relative_crowding.iloc[-1][i], index_relative_crowding_change.iloc[-1][i]), xytext=(index_relative_crowding.iloc[-1][i], index_relative_crowding_change.iloc[-1][i]))  # 这里xy是需要标记的坐标，xytext是对应的标签坐标
    plt.xlabel('拥挤度相对水平')
    plt.ylabel('拥挤度相对水平近一月变化')
    plt.xlim([-1.5, 1.5])
    plt.ylim([-1.5, 1.5])
    plt.hlines(y=0, xmin=-1.5, xmax=1.5, linestyles='-', color='#959595')
    plt.vlines(x=0, ymin=-1.5, ymax=1.5, linestyles='-', color='#959595')
    plt.gca().xaxis.set_major_formatter(FuncFormatter(to_100percent))
    plt.gca().yaxis.set_major_formatter(FuncFormatter(to_100percent))
    plt.title('先进制造各板块拥挤度情况')
    plt.tight_layout()
    plt.savefig('{0}拥挤度.png'.format(data_path))




import pandas as pd
from WindPy import w
import re
from hbshare.quant.CChen.func import generate_table
from datetime import timedelta, datetime
from time import sleep
from hbshare.quant.CChen.data.wind_cons import sql_l_futures, sql_l_bar, sql_l_stk, sql_l_edb


class WindApi(object):

    def __init__(self):
        self.api = w
        self.api.start()

    def get_com_fut_contracts(self, date):
        com = self.api.wset("sectorconstituent", "date=" + date.strftime('%Y-%m-%d') + ";sectorid=1000015512000000")
        com_df = pd.DataFrame(com.Data).T.rename(columns={0: 't_date', 1: 'wind_code', 2: 'sec_name'})
        com_df['t_date'] = pd.to_datetime(com_df['t_date']).dt.date
        return com_df

    def get_fin_fut_contracts(self, date):
        fin = self.api.wset("sectorconstituent", "date=" + date.strftime('%Y-%m-%d') + ";sectorid=a599010101000000")
        fin_df = pd.DataFrame(fin.Data).T.rename(columns={0: 't_date', 1: 'wind_code', 2: 'sec_name'})
        fin_df['t_date'] = pd.to_datetime(fin_df['t_date']).dt.date
        return fin_df

    def get_fut_data(self, date, com=True, fin=True):
        if com:
            com_contracts = self.get_com_fut_contracts(date=date)
        else:
            com_contracts = pd.DataFrame()

        if fin:
            fin_contracts = self.get_fin_fut_contracts(date=date)
        else:
            fin_contracts = pd.DataFrame()

        contracts = pd.concat([com_contracts, fin_contracts])

        contracts_str = str(contracts['wind_code'].tolist())
        contracts_str = contracts_str.replace('[', '').replace(']', '').replace('\'', '')

        properties = 'dlmonth, pre_close, pre_settle, open, high, low, close, settle, volume, amt, oi, oi_chg'

        data = self.api.wss(
            contracts_str,
            properties,
            'tradeDate=' + date.strftime('%Y%m%d')
        )

        data_df = pd.DataFrame(data.Data, index=data.Fields, columns=data.Codes).T
        data_df['t_date'] = date

        return data_df

    def to_sql_stk_index_fut(self, db_path, table, sql_info, end_date, database):
        generate_table(
            database=database,
            table=table,
            generate_sql=sql_l_futures,
            sql_ip=sql_info['ip'],
            sql_user=sql_info['user'],
            sql_pass=sql_info['pass'],
            table_comment='Wind期货连续合约行情'
        )
        print(table + ' generated')

        futures_list = pd.read_csv('wind_fut_list.csv')

        properties = "open, high, low, close, volume, amt, trade_hiscode, settle, oi, oiamount"

        for i in range(len(futures_list)):
            name = futures_list['name'][i]
            symbol = futures_list['code'][i]
            underlying = futures_list['underlying'][i]
            # product = nh_index_list['product'][i]

            print('Wind ' + name)

            data_exists = pd.read_sql_query(
                'select * from ' + table + ' where symbol="' + symbol.lower() + '" order by t_date', db_path
            )

            if len(data_exists) > 0:
                data_end_date = data_exists['t_date'].tolist()[-1]
                data_end_close = data_exists['close'].tolist()[-1]
                print(
                    '\tExisting ' + data_exists['t_date'].tolist()[0].strftime('%Y/%m/%d')
                    + ' - ' + data_exists['t_date'].tolist()[-1].strftime('%Y/%m/%d')
                )
                if data_end_date >= end_date.date():
                    print('\tData exists')
                    continue
                else:
                    start_date = (data_exists['t_date'].tolist()[-1] + timedelta(days=1))
            else:
                print('\tExisting None')
                start_date = datetime(1990, 1, 1)
                data_end_close = None

            print('\t' + str(i + 1) + '/' + str(len(futures_list)))
            data_raw = self.api.wsd(
                symbol,
                properties,
                start_date,
                end_date
            )
            underlying_raw = self.api.wsd(
                underlying,
                'close',
                start_date,
                end_date
            )
            if len(data_raw.Data[0]) == 0:
                print('\tNo new data')
                continue

            if data_raw.ErrorCode != 0:
                print(data_raw.ErrorCode)
                print('Wind Error')
                break

            data_new = pd.DataFrame(
                {
                    't_date': data_raw.Times,
                    'open': data_raw.Data[0],
                    'high': data_raw.Data[1],
                    'low': data_raw.Data[2],
                    'close': data_raw.Data[3],
                    'volume': data_raw.Data[4],
                    'amount': data_raw.Data[5],
                    'code': data_raw.Data[6],
                    'settle': data_raw.Data[7],
                    'oi': data_raw.Data[8],
                    'oi_amount': data_raw.Data[9]
                }
            )

            data_new = data_new[data_new['close'] > 0].reset_index(drop=True)
            data_new['symbol'] = symbol
            data_new['product'] = re.findall(r'^[a-zA-Z]+', symbol)[0]
            data_new['pre_close'] = data_new['close'].shift(1)
            data_new.loc[0, 'pre_close'] = data_end_close

            underlying_data = pd.DataFrame(
                {
                    't_date': underlying_raw.Times,
                    'underlying': underlying_raw.Data[0]
                }
            )

            data_new = data_new.merge(underlying_data, on=['t_date'])
            data_new['basis'] = data_new['underlying'] - data_new['close']

            contracts = data_new[['code']].drop_duplicates().reset_index(drop=True)
            contracts_info = w.wss(
                str(tuple(contracts['code'])).replace('(', '').replace(',)', '').replace('\'',''),
                'lasttrade_date'
            )
            contracts_info = pd.DataFrame(
                {
                    'code': contracts_info.Codes,
                    'delist_date': contracts_info.Data[0]
                }
            )
            data_new = data_new.merge(contracts_info, on=['code'])

            print('\tLoading ' + data_new['t_date'].tolist()[0].strftime('%Y/%m/%d')
                  + ' - ' + data_new['t_date'].tolist()[-1].strftime('%Y/%m/%d'))
            data_new.to_sql(table, db_path, if_exists='append', index=False)
            print('\tto sql')
            sleep(1)

    def to_sql_fut_index(self, db_path, table, sql_info, end_date, database, code_df, freq=''):
        generate_table(
            database=database,
            table=table,
            generate_sql=sql_l_bar,
            sql_ip=sql_info['ip'],
            sql_user=sql_info['user'],
            sql_pass=sql_info['pass'],
            table_comment='期货指数行情'
        )
        print(table + ' generated')

        properties = "open, high, low, close, volume, amt, oi, oiamount"
        for i in range(len(code_df)):
            # name = code_df['name'][i]
            # exchange = code_df['exchange'][i]
            code = code_df['code'][i]
            product = code_df['product'][i]

            print('Wind ' + code + ' ' + str(product))

            data_exists = pd.read_sql_query(
                'select * from ' + table + ' where code="' + code.lower() + '" order by t_date',
                db_path
            )

            if len(data_exists) > 0:
                data_end_date = data_exists['t_date'].tolist()[-1]
                data_end_close = data_exists['close'].tolist()[-1]
                print(
                    '\tExisting ' + data_exists['t_date'].tolist()[0].strftime('%Y/%m/%d')
                    + ' - ' + data_exists['t_date'].tolist()[-1].strftime('%Y/%m/%d')
                )
                if data_end_date >= end_date.date():
                    print('\tData exists')
                    continue
                else:
                    start_date = (data_exists['t_date'].tolist()[-1] + timedelta(days=1))
            else:
                print('\tExisting None')
                start_date = datetime(1990, 1, 1)
                data_end_close = None

            print('\t' + str(i + 1) + '/' + str(len(code_df)))

            if freq.lower() in ['w', 'week', 'fund_stats']:
                data_raw = self.api.wsd(
                    code,
                    properties,
                    start_date,
                    end_date,
                    options="Period=W"
                )
            elif freq.lower() in ['month', 'monthly']:
                data_raw = self.api.wsd(
                    code,
                    properties,
                    start_date,
                    end_date,
                    options="Period=M"
                )
            else:
                data_raw = self.api.wsd(
                    code,
                    properties,
                    start_date,
                    end_date
                )
            if len(data_raw.Data[0]) == 0:
                print('\tNo new data')
                continue

            if data_raw.ErrorCode != 0:
                print(data_raw.ErrorCode)
                print('Wind Error')
                break

            data_new = pd.DataFrame(
                {
                    't_date': data_raw.Times,
                    'open': data_raw.Data[0],
                    'high': data_raw.Data[1],
                    'low': data_raw.Data[2],
                    'close': data_raw.Data[3],
                    'volume': data_raw.Data[4],
                    'amount': data_raw.Data[5],
                    'oi': data_raw.Data[6],
                    'oi_amount': data_raw.Data[7],
                }
            )
            data_new = data_new[data_new['close'] > 0].reset_index(drop=True)
            data_new['code'] = code
            data_new['product'] = product
            # data_new['name'] = name
            # data_new['exchange'] = exchange
            data_new['pre_close'] = data_new['close'].shift(1)
            data_new.loc[0, 'pre_close'] = data_end_close

            data_new['chg'] = data_new['close'] - data_new['pre_close']
            data_new['pct_chg'] = data_new['chg'] / data_new['pre_close']

            # print(data_new)

            # data_new['t_date'] = pd.to_datetime(data_new['t_date']).dt.date
            print('\tLoading ' + data_new['t_date'].tolist()[0].strftime('%Y/%m/%d')
                  + ' - ' + data_new['t_date'].tolist()[-1].strftime('%Y/%m/%d'))
            data_new.to_sql(table, db_path, if_exists='append', index=False)
            print('\tto sql')
            sleep(1)

    def to_sql_a_shares(self, start_date, end_date, db_path, table, sql_info, database):
        used_amount_all = 0

        generate_table(
            database=database,
            table=table,
            generate_sql=sql_l_stk,
            sql_ip=sql_info['ip'],
            sql_user=sql_info['user'],
            sql_pass=sql_info['pass'],
            table_comment='A股日行情'
        )
        print(table + ' generated')

        fields = 'open,high,low,close,volume,amt,adjfactor,free_turn_n,trade_status,riskwarning,share_totaltradable'

        date_list = self.api.wsd('000001.SH', 'close', start_date.strftime('%Y-%m-%d'), end_date.strftime('%Y-%m-%d')).Times

        for s in date_list:
            print(s)

            data = self.api.wset("sectorconstituent", "date=" + s.strftime('%Y-%m-%d') + ";sectorid=a001010100000000")

            stocks_pool = data.Data[1]
            stocks_name = data.Data[2]

            name_df = pd.DataFrame(
                {
                    'code': stocks_pool,
                    'name': stocks_name
                }
            )

            data_local = pd.read_sql_query(
                'select * from ' + table + ' where t_date=' + s.strftime('%Y%m%d'), db_path
            )

            data_diff_local = list(set(stocks_pool).difference(set(data_local['code'])))

            stock_data_raw = self.api.wss(
                codes=str(data_diff_local).replace('[', '').replace(']', '').replace('\'', ''),
                fields=fields,
                options='tradeDate=' + s.strftime('%Y%m%d')
            )

            if stock_data_raw.ErrorCode != 0:
                print(stock_data_raw.ErrorCode)
                print('Wind Error')
                quit()

            used_amount = len(stock_data_raw.Data) * len(stock_data_raw.Codes)
            used_amount_all += used_amount
            stock_data = pd.DataFrame(
                {
                    't_date': s,
                    'code': stock_data_raw.Codes,
                    'open': stock_data_raw.Data[0],
                    'high': stock_data_raw.Data[1],
                    'low': stock_data_raw.Data[2],
                    'close': stock_data_raw.Data[3],
                    'volume': stock_data_raw.Data[4],
                    'amount': stock_data_raw.Data[5],
                    'adjFactor': stock_data_raw.Data[6],
                    'freeTurn': stock_data_raw.Data[7],
                    'tdState': stock_data_raw.Data[8],
                    'stState': stock_data_raw.Data[9],
                    'freeShare': stock_data_raw.Data[10]
                }, index=range(len(stock_data_raw.Codes))
            ).merge(name_df, on='code', how='left')

            stock_data.to_sql(table, db_path, if_exists='append', index=False)
            print(
                # s.strftime('%Y-%m-%d') +
                '\t ' + str(len(stock_data)) + ' stk done'
                + '\n\tused amount this time: ' + str(used_amount)
                + '\n\tused amount this all: ' + str(used_amount_all)
            )

    def get_edb_data(self, start_date, end_date, db_path, table, sql_info, database, indicators):
        generate_table(
            database=database,
            table=table,
            generate_sql=sql_l_edb,
            sql_ip=sql_info['ip'],
            sql_user=sql_info['user'],
            sql_pass=sql_info['pass'],
            table_comment='Wind EDB数据'
        )
        print(table + ' generated')
        start_date0 = start_date

        for i in range(len(indicators)):
            print(indicators['指标名称'][i])
            data_exists = pd.read_sql_query(
                'select * from ' + table
                + ' where code=\'' + indicators['指标ID'][i]
                + '\' order by t_date',
                db_path
            )

            if len(data_exists) > 0:
                data_end_date = data_exists['t_date'].tolist()[-1]
                print(
                    '\tExisting ' + data_exists['t_date'].tolist()[0].strftime('%Y/%m/%d')
                    + ' - ' + data_exists['t_date'].tolist()[-1].strftime('%Y/%m/%d')
                )
                if data_end_date >= end_date:
                    print('\t\tData exists')
                    continue
                else:
                    start_date = (data_exists['t_date'].tolist()[-1] + timedelta(days=1))
            else:
                print('\tExisting None')
                start_date = start_date0

            data_raw = self.api.edb(
                indicators['指标ID'][i],
                start_date.strftime('%Y-%m-%d'), end_date.strftime('%Y-%m-%d')
            )
            if data_raw.ErrorCode != 0:
                print('\tWind error: ' + str(data_raw.ErrorCode))
                continue

            df = pd.DataFrame(
                {
                    'code': data_raw.Codes[0],
                    't_date': data_raw.Times,
                    'close': data_raw.Data[0]
                }
            ).dropna()
            df = df[df['t_date'] >= start_date]
            if len(df) > 0:
                df.to_sql(table, db_path, if_exists='append', index=False)
                print(
                    '\t\t更新 '
                    + df['t_date'].tolist()[0].strftime('%Y-%m-%d') + ' - '
                    + df['t_date'].tolist()[-1].strftime('%Y-%m-%d')
                )
            else:
                print('\t\t 无更新')

        # print()


import numpy as np
import pandas as pd
from hbshare.fe.XZ import db_engine as dbeng
from hbshare.fe.XZ import functionality
from sklearn import  preprocessing as pp



util=functionality.Untils()
hbdb=dbeng.HBDB()
fre_map={'M':'2101','Q':'2103','HA':'2106','A':'2201','2A':'2202','3A':'2203'}
localdb = dbeng.PrvFunDB().engine

class Multi_factor_model:

    def __init__(self,factor_list,fre):

        self.factor_list=factor_list
        self.fre=fre

        factor_df=pd.DataFrame()
        count=0
        for factor_name in factor_list:

            raw_factor_data = self.factor_loader(factor_name)

            raw_factor_data['month'] = raw_factor_data['date'].astype(str).str[4:6]
            raw_factor_data['year_month']=raw_factor_data['date'].astype(str).str[0:6]

            if (fre == 'Q'):
                raw_factor_data = raw_factor_data.loc[raw_factor_data['month'].isin(['03', '06', '09', '12'])]
            elif (fre == 'HA'):
                raw_factor_data = raw_factor_data.loc[raw_factor_data['month'].isin(['06', '12'])]
            elif (fre == 'A'):
                raw_factor_data = raw_factor_data.loc[raw_factor_data['month'].isin(['12'])]

            if (count == 0):
                factor_df = raw_factor_data.drop(['month'],axis=1).copy()
            else:
                factor_df = pd.merge(factor_df
                                     , raw_factor_data.drop(['date','month'],axis=1), how='inner', on=['jjdm', 'year_month'])
            count += 1

        self.raw_factor_data=factor_df.drop('year_month',axis=1)


    @staticmethod
    def factor_loader(factor_name):

        if('25%' in factor_name):

            rolling=factor_name.split('25%')[1]

            sql="SELECT * from factor{}_rank_quantile".format(rolling)
            factor_date=pd.read_sql(sql,con=localdb)[['jjdm','25%','date']].rename(columns={'25%':factor_name})

        elif (factor_name == 'year_logret'):

            sql="select jjdm,year_logret,date from factor_year_ret where year_logret is not null "
            factor_date=pd.read_sql(sql, con=localdb)

        elif (factor_name == 'jjzzc'):

            sql="select jjdm,jjzzc,date from factor_size  "
            factor_date=pd.read_sql(sql, con=localdb)

        elif ('info_ratio' in factor_name):

            rolling=factor_name.split('info_ratio')[1].split('_')[1]

            sql="select jjdm,info_ratio as {0},date from factor_infor_ratio where zblb={1} ".format(factor_name,rolling)
            factor_date=pd.read_sql(sql, con=localdb)
            factor_date['date']=factor_date['date'].astype(str)

        elif ('sortino' in factor_name):

            rolling=factor_name.split('sortino')[1].split('_')[1]

            sql="select jjdm,sortino as {0},date from factor_sortino where zblb={1} ".format(factor_name,rolling)
            factor_date=pd.read_sql(sql, con=localdb)
            factor_date['date']=factor_date['date'].astype(str)

        elif ('ext_ret' in factor_name):

            from hbshare.fe.mutual_analysis import nav_based as nbs

            sr = nbs.Scenario_return()
            factor_date = sr.factorlize_ret(factor_name)[['jjdm','date']+[factor_name]]
            factor_date['date']=factor_date['date'].astype(str)

        else:
            print('input factor name is not included in the model by now ')
            raise  Exception

        return  factor_date

    def factor_correlation_check(self):

        factor_df=self.raw_factor_data.copy()
        factor_df.drop('jjdm',axis=1,inplace=True)

        corr_res=factor_df.groupby('date').corr()

        return  corr_res.reset_index().groupby('level_1').mean()

    def get_the_combined_factors(self,method,factor_list=None,factor_symbol=None):

        if(factor_list is None):
            factor_list=self.factor_list

        factor_df = self.raw_factor_data[['jjdm', 'date'] + factor_list]
        factor_df[factor_list] = pp.scale(factor_df[factor_list])


        if(method=='avg'):
            if(factor_symbol is None):
                factor_df['new_factor']=factor_df[factor_list].mean(axis=1)
            else:
                for j in range(len(factor_list)):
                    factor_df[factor_list[j]]=factor_df[factor_list[j]]*factor_symbol[j]
                factor_df['new_factor'] = factor_df[factor_list].mean(axis=1)

        elif(method=='ir'):

            from hbshare.fe.factor_analysis import  single_factor_analysis as sfa
            factor_df['new_factor']=0
            factor_df['total_ir'] = 0


            for factor in factor_list:

                ic_history=sfa.single_factor_ic(factor_df[['jjdm','date']+[factor]],factor
                                           ,fre=self.fre,pool_size=20,fund_flag=True,grouping_flag=False)
                ic_history['std']=ic_history.rolling(120,4).std()['ic']
                ic_history['mean'] = ic_history.rolling(120,4).mean()['ic']
                ic_history['ir']=ic_history['mean']/ic_history['std']
                # ic_history['ir']=[np.max([x,0]) for x in ic_history['ir']]

                factor_df=pd.merge(factor_df
                                   ,ic_history[['date','ir']].rename(columns={'ir':factor+"_ir"}),how='left',on='date')

                factor_df['new_factor']=factor_df['new_factor']+factor_df[factor]*factor_df[factor+"_ir"]
                factor_df['total_ir'] = factor_df['total_ir'] + factor_df[factor + "_ir"].abs()

            factor_df['new_factor']=factor_df['new_factor']/factor_df['total_ir']
            factor_df=factor_df[(factor_df['total_ir'].notnull())|(factor_df['date']==factor_df['date'].max())]

        else:
            print('the input method is not suppoered ')
            raise  Exception


        return factor_df[['jjdm','date','new_factor']]


if __name__ == '__main__':


    mfm=Multi_factor_model(['25%_year','25%_half_year','year_logret'],fre='Q')
    #corr=mfm.factor_correlation_check()




    print('Done')